use std::{collections::VecDeque, fmt::Display};

use bitvec::prelude::*;
use log::{error, warn};

use super::{wire::Request, Coord, Direction};

#[derive(Debug, PartialEq, Eq, Clone)]
pub struct Board {
    width: u8,
    height: u8,
    hazard_damage: u8,
    food_spawn_chance: u8,
    min_food: u16,
    turn: u32,
    food: BitBox,
    hazard: BitBox,
    free: BitBox,
    snakes: Vec<Snake>,
}

impl From<&Request> for Board {
    fn from(value: &Request) -> Self {
        let width = value.board.width;
        let height = value.board.height;
        let fields = usize::from(width) * usize::from(height);
        let mut board = Self {
            width,
            height,
            hazard_damage: value.game.ruleset.settings.hazard_damage_per_turn,
            food_spawn_chance: value.game.ruleset.settings.food_spawn_chance,
            min_food: value.game.ruleset.settings.minimum_food,
            turn: value.turn,
            food: bitbox![0; fields],
            hazard: bitbox![0; fields],
            free: bitbox![1; fields],
            snakes: Vec::with_capacity(value.board.snakes.len()),
        };

        for &food in &value.board.food {
            let index = usize::from(board.coord_to_linear(food));
            board.food.set(index, true);
        }

        for &hazard in &value.board.hazards {
            let index = usize::from(board.coord_to_linear(hazard));
            board.hazard.set(index, true);
        }

        for (id, snake) in value.board.snakes.iter().enumerate() {
            for &tile in &snake.body {
                let index = usize::from(board.coord_to_linear(tile));
                board.free.set(index, false);
            }
            let snake = Snake {
                body: snake.body.iter().copied().collect(),
                id: u8::try_from(id).unwrap_or(u8::MAX),
                health: snake.health,
            };
            board.snakes.push(snake);
        }

        board
    }
}

impl Display for Board {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(
            f,
            "{}x{} {}% ({}) {}dmg @ {}",
            self.width,
            self.height,
            self.food_spawn_chance,
            self.min_food,
            self.hazard_damage,
            self.turn
        )?;

        for y in (0..self.height).rev() {
            for x in 0..self.width {
                let tile = Coord { x, y };
                if self.snakes.iter().any(|snake| snake.head() == tile) {
                    write!(f, "H")?;
                    continue;
                }
                if self.snakes.iter().any(|snake| snake.tail() == tile) {
                    write!(f, "T")?;
                    continue;
                }

                let index = usize::from(self.coord_to_linear(tile));
                if !self.free[index] {
                    write!(f, "S")?;
                    continue;
                }
                if self.food[index] {
                    write!(f, "f")?;
                    continue;
                }
                if self.hazard[index] {
                    write!(f, "h")?;
                    continue;
                }

                write!(f, ".")?;
            }
            writeln!(f)?;
        }

        Ok(())
    }
}

impl Board {
    pub fn is_food(&self, tile: Coord) -> bool {
        let index = usize::from(self.coord_to_linear(tile));
        self.food[index]
    }

    pub fn is_hazard(&self, tile: Coord) -> bool {
        let index = usize::from(self.coord_to_linear(tile));
        self.hazard[index]
    }

    pub fn is_free(&self, tile: Coord) -> bool {
        let index = usize::from(self.coord_to_linear(tile));
        self.free[index]
    }

    pub fn valid_actions(&self, id: u8) -> impl Iterator<Item = Direction> + use<'_> {
        let head = self
            .snakes
            .binary_search_by_key(&id, |snake| snake.id)
            .ok()
            .map(|index| self.snakes[index].head());
        if head.is_none() {
            warn!(
                "Asked for an action for a snake that doesn't exist: {id} not in {:?}",
                self.snakes
            );
        }
        enum_iterator::all::<Direction>()
            .filter_map(move |direction| {
                head.and_then(|head| head.apply(direction))
                    .map(|tile| (direction, tile))
            })
            .filter(|(_, tile)| tile.x < self.width && tile.y < self.height)
            .filter(|(_, tile)| self.is_free(*tile))
            .map(|(direction, _)| direction)
    }

    fn coord_to_linear(&self, coord: Coord) -> u16 {
        u16::from(coord.x) + u16::from(coord.y) * u16::from(self.width)
    }
}

#[derive(Debug, PartialEq, Eq, Clone)]
struct Snake {
    id: u8,
    health: u8,
    body: VecDeque<Coord>,
}

impl Snake {
    pub fn head(&self) -> Coord {
        self.body.front().copied().unwrap_or_else(|| {
            error!("Snake without a head: {self:?}");
            Coord { x: 0, y: 0 }
        })
    }

    pub fn tail(&self) -> Coord {
        self.body.back().copied().unwrap_or_else(|| {
            error!("Snake without a tail: {self:?}");
            Coord { x: 0, y: 0 }
        })
    }
}
